// Copyright 1995-2016 The OpenSSL Project Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef OPENSSL_HEADER_TLS_PRF_H
#define OPENSSL_HEADER_TLS_PRF_H

#include <openssl/base.h>  // IWYU pragma: export

#if defined(__cplusplus)
extern "C" {
#endif


// TLS PRF.
//
// The TLS PRF is defined in Section 5 of RFC 5246.


// CRYPTO_tls1_prf calculates `out_len` bytes of the TLS PRF, using `digest`,
// and writes them to `out`. It is defined in Section 5 of RFC 5246, acting on
// `secret_len` bytes of shared `secret`, `label_len` bytes of `label`,
// `seed1_len` bytes of `seed1` and `seed2_len` bytes of `seed2`. It returns one
// on success and zero on error.
OPENSSL_EXPORT int CRYPTO_tls1_prf(const EVP_MD *digest, uint8_t *out,
                                   size_t out_len, const uint8_t *secret,
                                   size_t secret_len, const uint8_t *label,
                                   size_t label_len, const uint8_t *seed1,
                                   size_t seed1_len, const uint8_t *seed2,
                                   size_t seed2_len);


#if defined(__cplusplus)
}  // extern C
#endif

#endif  // OPENSSL_HEADER_TLS_PRF_H
